
;
; Copyright 1999 by Konrad Rzeszutek 
; Date:    Sun 28th Feb 1999
; File:    8-puzzle.lsp
;
; Notes:
;
; Visualize a board 3x3 filled with numbers ranging from 1 to 8. The ninth is a
; blank spot  used so that the tiles can be moved up/down or left/right
; depending on the situation. The goal of the game is to achieve a state where
; all the numbers are in a specific order
;
; This 8 puzzle uses A* (graph extension) heuristic search and uses
; three diffrent heuristicsm functions:
;
; o brute-force which always gives  h(t)=0
; o number-out-of-place, which simply counts out of place items
; o city-block-distance which is based on Manhattan distance formula
;

; Prologue:
;
;  A "board" is a list of 8 integers and a '- (to symbolize blank spot)
;  first three items are the first row, the next three items are the second row,
;  and so on.
;
;  A "node" will be a generated symbol carrying the properties:
;    o  board           - the configuration of the puzzle
;    o  cost            - the "quality" of the node (currently not implemented)
;    o  g-value         - the depth of the node
;    o  parent-node     - parent node to this node
;    o  generating-move - move made from parent to reach this node

;
; The main module is
;
;   A*-search-of-8-puzzle 
;
; Which accepts the starting board and the goal board. It is based on A* search
; in trying to solve the puzzle. The main moduile will search for the goal-board
; using three diffrent heuristic functions: brute=force, out-of-place, and 
; manhattan-distance.

;
; This is used by h(t)
; 
(defvar *heuristicism*  '(manhattan-distance out-of-place brute-force))

;
; The various move commands
;

(defvar *move-list* '(move-right move-up move-left move-down))


; 9 element boards

(defvar *goal*   '(1 2 3 8 - 4 7 6 5))
(defvar *easy*   '(1 3 4 8 6 2 7 - 5))
(defvar *medium* '(2 8 1 - 4 3 7 6 5))
(defvar *hard*   '(2 8 1 4 6 3 - 7 5))
(defvar *s-hard* '(1 2 3 4 5 6 7 8 -))

;
; Constants
;
(defvar *level-size* 3)
(defvar *blank-spot* '-)

(defun all ()

  (let
      ((h (copy-list  *heuristicism*))); list of heuretistic functions

    (dolist (function  h)
      (setq  *heuristicism* (list function)); yes, we are overwritting the *heuristic* list, but will later have the original.
      (princ "**EASY**:")
      (time (A*-search-of-8-puzzle  *easy* *goal*))

      (princ "**MEDIUM**:")
      (time (A*-search-of-8-puzzle *medium* *goal*))

      (princ "**HARD**:")
      (time (A*-search-of-8-puzzle *hard* *goal*))

      )
    (setq  *heuristicism* '(manhattan-distance))
    (princ "**SUPER-HARD**:")
    (time (A*-search-of-8-puzzle *s-hard* *goal*)) ; this takes f-o-r-e-v-e-r
    (setq *heuristicism* h)            ; put the original data back
    )
  )

; +==========================================================================+
; | Main module                                                              |
; |                                                                          |
; +==========================================================================+
;
(defun A*-search-of-8-puzzle (start-board goal-board)

  (dolist (h-of-t *heuristicism*)

    (format t "~%~%Finding the path for:~%~%")
    (output-board start-board)
    (format t "~%To goal board: ~%~%")
    (output-board goal-board)
    (format t "~%Using ~A heurestic function. ~%~%Pondering ." h-of-t)
    
     (let(
	  
	  (Open                           ; open list with a singleton
	   (list (make-node start-board   ; board configuration
			    0             ; cost of node
			    1             ; g-value of node
			    nil           ; no parents
			    nil           ; and no moves
			    )))
		 (engulfed-nodes 1)       ; mnaim, mniam, also know as diff boards
		 first-node               ; from Open
		 next-board               ; generated from first-node board configuration
		 children-list            ; generated from first-node
		 child-node               ; node based on next-board
		 (found-goal nil)         ; for exiting the loop
		 goal-node                ; for output routine
		 (Closed nil)             ; passed nodes
		 )


       (loop
         (setq first-node (pop Open))      ; Open shortened as side effect
         (if (or found-goal (null first-node))
             ; then
             (return))                     ; goal or Open empty
           ; else go on
         (setq children-list nil)
	
         (dolist (move *move-list*)

           (setq next-board (funcall move (board-of first-node)))
	   ; next-board will either have result of the one of the
           ; move modules-commands or null (if the move is unmakeable)

           (when next-board
             (setq child-node (make-node 
                               next-board    ; 
                               0             ; cost of all nodes will be 0
                               (+ (g-value-of first-node) (+ (cost-of first-node) 0) 1)
                                             ; g(child) = g(parent) + 1+cost-of-arc(child-parent)
                               first-node move))  ; parent and the move
	     
	     (when (is-goal-board child-node goal-board)
	       (setq found-goal t)
	       (setq goal-node child-node)
	       (return)
	       )	     
	     (unless (is-node-in child-node Closed) ; check against duplicates
	       (incf engulfed-nodes 1)
	       (push child-node children-list)     ; update good moves, not duplicates
	       )
	     )
	   )

	 (setq Open (nconc Open children-list))
	 (setq Open (sort Open  #'(lambda (x y) 
				    (when (<= 
					   (+ (g-value-of x) (funcall h-of-t (board-of x) goal-board))
					   (+ (g-value-of y) (funcall h-of-t (board-of y) goal-board))
					   ) t))))
					; ^^ - sorts the Open list in increasing order of f(n).
					; f(n) = g(n) + h(n)
	 (push first-node Closed)       ; used to find duplicates
	 ) ; end of A* loop

       (if found-goal
	   (output-solution-path goal-node engulfed-nodes)
	 (output-no-solution))
       )
     )
  )



; +==========================================================================+
; | Display routines                                                         |
; |                                                                          |
; +==========================================================================+
;

; 
; The path to the solution
;
(defun output-solution-path (goal-node encountered-boards)
    
  (let(
       (goal-path nil)  ; list of visited nodes
       )
    (do*(
	 (Q goal-node (parent-node-of Q)) ; its organized as a stack, need to make it a list
	 )
	((eql nil Q) nil)
      (setq goal-path (nconc (list Q) goal-path)) ; put them in reversal order in the list
      )
    
    (dolist (Q goal-path)
      
      (format t "~%~%~A~%~%" (generating-move-of Q)) ; the move
      (format t "~A" (output-board (board-of Q)))    ; the board
      )
    
    (format t "~%The solution path has length = ~A~%The number of diffrent boards encountered = ~A~%~%" (length goal-path) encountered-boards)
  
  )
)


(defun output-no-solution () 
  
  (format t "~%Sorry, no possible path!")
  )

;
; Display the board
;
(defun output-board (board)
  
  (do*(
       (column 0 (+ column 1)) ; column of the board
       (row    0)              ; row of the board
       )
      ((= row *level-size*) nil) ; when row is equal to the last row
    (princ (nth (+ column (* row *level-size*)) board)) (princ " ") ; row*column of the board
    (when (= column (- *level-size* 1)) ; column is by the end
      (incf row 1)
      (setf column -1)
      (terpri))
    )
  (princ " ")
  )

      
    


; +==========================================================================+
; | Heuristic functions
; +==========================================================================+

;
; h(t) == 0
;
(defun brute-force (board goal-board)

  ; does exactly nothing, great heuristic!
  0
  
)
;
; h(t) >= 0
;
; Counts out of place elements on the board as compared to the goal-board
;
(defun out-of-place (board goal-board)

  (let(
       (counter 0)         ; index on the goal-board
       (out_of_place 0)    ; amount of out of place elements
       )
    (dolist (element board)
      (if (not (eql element (nth counter goal-board))) 
	  ; then
	   (incf out_of_place 1)
	   )	
      (incf counter 1)
      )
    out_of_place
    )
)

;
; h(t) >= 0
; Manhattan distance, also know as 'city block-distance' adds the number of rows
; and columns that it needs to move in order to occupy its intended location.
;
; Formula:
;  | r - r'| + | c - c'|
;
(defun manhattan-distance (board goal-board)

  (let(
       (city_block_distance 0)  ; results of added rows and columns
       )

    (do(
	(i 0 (+ i 1)) ; index through the board
	)
	; exit test:
	( (= i (expt *level-size* 2)) city_block_distance) ; is i==3^2 ?

	(if (not (eql (nth i board) (nth i goal-board))) ; if in wrong location
	    ; then
	    (setq city_block_distance (+ city_block_distance 
					  
					  (abs (- (row-is-of (+ i 1)) (row-is-of (position-of (nth i goal-board) board)))) ; |r - r'|
					  (abs (- (column-is-of (+ i 1)) (column-is-of (position-of (nth i goal-board) board)))) ; | c - c'|
					  ))
	  ) ; endif  
	
	)
    city_block_distance
    )
  )
  



; +==========================================================================+
; | Move commmands                                                           |
; |                                                                          |
; | Each returns the new board configuration from the move,                  |
; | but returns nil if the move is illegal                                   |
; |                                                                          |
; +==========================================================================+
;

(defun move-left (board)
  
  (let
      ( (blank-spot (- (position-of-blank-spot-in board) 1)); pos. of the blank spot
	(new-board (copy-list board))                       ; we won't maipulate on the old board
	)
    (cond
     ((> (column-is-of (+ blank-spot 1)) 1)      
      (setf (nth  blank-spot  new-board) (nth (- blank-spot 1)  new-board)) ; copy from n-1 to n
      (setf (nth (- blank-spot 1) new-board) *blank-spot*)               ; copy *blank-spot to n-1
      new-board ; returns new board
      )
     )
    )
  )


  
(defun move-right (board)
  
  (let
      ( (blank-spot (- (position-of-blank-spot-in board) 1)); pos. of the spot
	(new-board (copy-list board))                       ; we won't maipulate on the old board
	)
    (cond
     ((< (column-is-of (+ blank-spot 1)) *level-size*) ; in case we will make the board bigger
      (setf (nth  blank-spot  new-board) (nth (+ blank-spot 1)  new-board)) ; copy from n+1 to n
      (setf (nth (+ blank-spot 1) new-board) *blank-spot*)               ; copy *blank-spot to n+1
       new-board ; returns new board
      )
     )
    )
  )  
     
(defun move-up (board)
  
  (let
      ( (blank-spot (position-of-blank-spot-in board)); pos. of the spot
	(new-board (copy-list board))                 ; we won't maipulate on the old board
	)
    (cond
     ((> (row-is-of blank-spot) 0)
      (setf (nth  (- blank-spot 1) new-board) (nth (- blank-spot (+ *level-size* 1))  new-board)) ; copy from n-*level-size* (above row) to n
      (setf (nth (- blank-spot (+ *level-size* 1)) new-board) *blank-spot*)               ; copy *blank-spot to n-*level-size*
       new-board ; returns new board
      )
     )
    )
  )  

 
(defun move-down (board)
  
  (let
      ( (blank-spot (position-of-blank-spot-in board)); pos. of the spot
	(new-board (copy-list board))                 ; we won't maipulate on the old board
	)
    (cond
     ((< (row-is-of blank-spot) (- *level-size* 1))
      (setf (nth  (- blank-spot 1) new-board) (nth (+ blank-spot (- *level-size* 1))  new-board)) ; copy from n+*level-size* (bottom row) to n (blank-spot)
      (setf (nth (+ blank-spot (- *level-size* 1)) new-board) *blank-spot*)               ; copy *blank-spot to n+*level-size*
      new-board  ; returns new board
      )
     )
    )
  )  

; +==========================================================================+
; | Various math routines                                                    |
; +==========================================================================+
;

;
; Calculates the column from the index
;
(defun column-is-of (index)

  (- index (* *level-size* (row-is-of index)))
)


;
; Calculates the row from the index
;
(defun row-is-of (index)

  (truncate (/ (- index 1) *level-size*))
)

; +==========================================================================+
; | Various board manipulation tools                                         |
; +==========================================================================+


(defun position-of-blank-spot-in (board)
 
  (let 
      ((index 0))                               ; location of the blank spot
    (dolist (elements board)
      (if (eql *blank-spot* (nth index board))
	  (return (+ index 1)))                 ; found it
      (incf index 1)
      )
    )
  )

;
; Finds the position of requested 'element' in the 'board'
;

(defun position-of (element board)

  ; The heuristic functions will only use this.

  (do(
       (i 0 (+ 1 i)) ; index on the board
       )
      ( (eql element (nth i board)) (+ i 1))  ; heurestic functions start at element 1, not 0
    )
  )

(defun same-board-configuration (board-one board-two)

  (do(
      (i 0 (+ 1 i)) ; index off the board-one
      (sum 0 )      ; sum of all the positions on the board
      )
      (  (= i (expt *level-size* 2)) (= sum i))
    
    (if (eql (nth i board-one) (nth i board-two))
	(incf sum 1)
      )
    )
  )

; +==========================================================================+
; | Tree manipulation functions
; +==========================================================================+
; 
;


;
; Returns 'node' if it exists in the same configuration in the
; space L, otherwise returns nil
;
(defun is-node-in (node L)
  
  (cond
   ( (null L)  ; void
     nil)
   ( (atom L)  ; 
     (same-board-configuration (board-of node) (board-of L)))     
   ( (listp L) ; ...dive in.
     (or (is-node-in node (first L)) (is-node-in  node (rest L))))
   )
  )


; +==========================================================================+
; | Node  manipulation functions                                             |
; +==========================================================================+
; 


;
; Creates the nodes for the list.
;

(defun make-node (board cost g-value parent-node generating-move )


  (let ((new-symbol (intern (symbol-name (gensym))))
	)
    (setf (get new-symbol 'board) board)
    (setf (get new-symbol 'g-value) g-value)
    (setf (get new-symbol 'cost) cost)
    (setf (get new-symbol 'parent-node) parent-node)
    (setf (get new-symbol 'generating-move) generating-move)
    new-symbol
    )
)

;
; Is this state the goal state?
;
(defun is-goal-board (state goal-board)

  (same-board-configuration  goal-board  (board-of state))

  )


;The "selectors":

(defun cost-of (node)

  (get node 'cost))

(defun g-value-of (node)
  
  (get node 'g-value))

(defun board-of (node)

  (get node 'board))

(defun child-list-of (node)
  
  (get node 'child-list))

(defun parent-node-of (node)

  (get node 'parent-node))

(defun generating-move-of (node)

  (get node 'generating-move))
